'''A Mock S3 Connection'''


from .bucket import Bucket
from .exceptions import S3ResponseError


class S3Connection(object):
    '''An S3 Connection. With buckets and everything'''
    def __init__(self):
        self._buckets = {}

    def create_bucket(self, bucket):
        '''Create a bucket'''
        if not bucket in self._buckets:
            self._buckets[bucket] = Bucket(bucket)

    def delete_bucket(self, bucket):
        '''Delete a bucket'''
        obj = self._buckets.get(bucket)
        if obj and not obj.empty():
            raise S3ResponseError('Bucket %s not empty' % bucket)
        self._buckets.pop(bucket, None)

    def get_bucket(self, bucket):
        '''Get a bucket'''
        obj = self._buckets.get(bucket)
        if not obj:
            raise S3ResponseError('Bucket %s does not exist' % bucket)
        return obj

    def get_all_buckets(self):
        '''Get ALL the buckets'''
        return self._buckets.values()

    def lookup(self, bucket):
        '''Get a bucket, or None if it doesn't exist'''
        return self._buckets.get(bucket, None)



'''Mock s3 access with the local filesystem'''

import os
import shutil


class Mock(object):
    '''Mocking base'''
    atts = []

    def __init__(self, obj):
        self._obj = obj
        self._original = {}

    def mock(self, obj, atts):
        '''Mock the object's attributes with this object's attributes'''
        for att in atts:
            self._original[att] = getattr(obj, att)
            setattr(obj, att, getattr(self, att))

    def unmock(self, obj, atts):
        '''Stop the mockery'''
        for att in atts:
            setattr(obj, att, self._original.pop(att))

    def start(self):
        '''Start mocking'''
        self.mock(self._obj, self.atts)

    def stop(self):
        '''Stop mocking'''
        self.unmock(self._obj, self.atts)

    def __enter__(self):
        self.start()

    def __exit__(self, typ, val, trace):
        self.unmock(self._obj, self.atts)
        if typ:
            raise typ, val, trace


class S3Key(object):
    '''Mock a key'''
    def __init__(self, bucket, name):
        self._bucket = bucket
        self._name = name
        self.path = os.path.join(bucket.path, name.rstrip('/'))

    def exists(self):
        '''Return true if this key exists'''
        return os.path.exists(self.path)

    @staticmethod
    def mkdir(path):
        '''Make the path and return it'''
        try:
            os.makedirs(os.path.dirname(path))
        except:
            pass
        return path

    def set_contents_from_string(self, string):
        '''Set the file's content from a string'''
        with open(self.mkdir(self.path), 'w+') as fout:
            fout.write(string)

    def get_contents_as_string(self):
        '''Get the contents as a string'''
        with open(self.path) as fin:
            return fin.read()


class S3Prefix(object):
    '''A Prefix'''
    def __init__(self, bucket, name):
        self.bucket = bucket
        self.name = name


class S3Bucket(object):
    '''Mock a bucket'''
    def __init__(self, base, name):
        self._base = base
        self._name = name
        self.path = os.path.join(base, name)

    def get_key(self, name):
        '''Get a key in this bucket, if it exists. Otherwise, None'''
        key = S3Key(self, name)
        if not key.exists():
            return None
        return key

    def list(self, prefix, separator=None):
        '''Iterate over keys beginning with this prefix, until separator'''
        paths = [(self._base, '')]
        found = []
        while paths:
            path, relative = paths.pop()
            if os.path.isdir(path):
                paths.extend(
                    (p, os.path.join(relative, p)) for p in os.listdir(path))
            else:
                found.append(path)

        paths = []
        for _, relative in found:
            if not relative.startswith(prefix):
                continue
            if separator:
                location = relative.find(separator)
                if location == -1:
                    paths.append(relative)
                else:
                    paths.append(relative[:location])
            else:
                paths.append(relative)

        # Now, uniquify them and turn them to objects
        paths = sorted(list(set(paths)))
        results = []
        for path in paths:
            if os.path.exists(os.path.join(self._base, path)):
                results.append(S3Key(self, path))
            else:
                results.append(S3Prefix(self, path))
        return results


class S3Connection(Mock):
    '''Mock the S3Connection for an S3Connection'''
    atts = ['get_bucket']

    def __init__(self, base, conn):
        Mock.__init__(self, conn)
        self._base = base

    def get_bucket(self, bucket):
        '''Get a bucket object'''
        return S3Bucket(self, bucket)


class S3po(Mock):
    '''Mock is a context manager that mocks s3 operations'''
    # These are the methods that this mocks on s3po
    atts = [
        'uploadString', 'uploadFile', 'downloadString', 'downloadFile', 'conn']

    def __init__(self, base, s3obj):
        Mock.__init__(self, s3obj)
        self._base = os.path.abspath(base)
        self.conn = S3Connection(self._base, s3obj.conn)

    def stop(self):
        '''Stop mocking'''
        # Stop the mocking for us and our connection
        Mock.stop(self)
        self.conn.stop()
        if os.path.exists(self._base):
            shutil.rmtree(self._base)

    ##########################################
    # Actual methods that it's mocking
    ##########################################
    def uploadString(self, bucket, key, string, *args, **kwargs):
        '''Upload the provided string'''
        S3Key(
            self.conn.get_bucket(bucket), key).set_contents_from_string(string)

    def uploadFile(self, bucket, key, path, *args, **kwargs):
        '''Upload the provided file'''
        with open(path) as fin:
            self.uploadString(bucket, key, fin.read())

    def downloadString(self, bucket, key, *args, **kwargs):
        '''Download the provided string'''
        return S3Key(self.conn.get_bucket(bucket), key).get_contents_as_string()

    def downloadFile(self, bucket, key, filename=None, *args, **kwargs):
        '''Download the provided file'''
        path = filename or os.path.dirname(key)
        with open(path, 'w+') as fout:
            fout.write(self.downloadString(bucket, key))



'''Test our Connection'''

import os
import shutil
import unittest
from cStringIO import StringIO

from s3po import Connection
from s3po.exceptions import UploadException, DownloadException


class ConnectionTest(unittest.TestCase):
    '''Test our connection's functionality'''
    tmpdir = 'test/tmp'

    def setUp(self):
        self.s3po = Connection()
        self.mock = self.s3po.mock()
        self.mock.start()
        self.s3po.conn.create_bucket('s3po')
        if not os.path.exists(self.tmpdir):
            os.mkdir(self.tmpdir)
        self.assertEqual(os.listdir(self.tmpdir), [])

    def tearDown(self):
        self.mock.stop()
        for path in os.listdir(self.tmpdir):
            path = os.path.join(self.tmpdir, path)
            if os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.remove(path)

    def test_upload(self):
        '''We should be able to upload a small key'''
        self.assertRaises(DownloadException, self.s3po.download, 's3po', 'foo')
        self.s3po.upload('s3po', 'foo', 'hello')
        self.assertEqual(self.s3po.download('s3po', 'foo'), 'hello')

    def test_multipart(self):
        '''If the key is large enough, we should do this as multipart'''
        self.s3po.multipart_chunk = 5 * 1024 * 1024
        data = 'hello' * self.s3po.multipart_chunk
        self.s3po.upload('s3po', 'foo', data)
        self.assertEqual(len(self.s3po.download('s3po', 'foo')), len(data))

    def test_short_download(self):
        '''If there is too little downloaded, throws an error'''
        self.s3po.upload('s3po', 'foo', 'bar')
        self.s3po.conn.get_bucket('s3po').get_key('foo')._size = 5
        self.assertRaises(DownloadException, self.s3po.download, 's3po', 'foo',
            retries=0)

    def test_short_upload(self):
        '''If we uploaded too little data, throws an error'''
        self.s3po.upload('s3po', 'foo', 'bar')
        self.s3po.conn.get_bucket('s3po').get_key('foo')._size = 5
        self.assertRaises(
            UploadException, self.s3po.upload, 's3po', 'foo', 'bar', retries=0)

    def test_upload_file(self):
        '''We should be able to upload a file'''
        self.s3po.upload('s3po', 'foo', StringIO('hello'))
        self.assertEqual(self.s3po.download('s3po', 'foo'), 'hello')

    def test_download_file(self):
        '''We should be able to download to a file'''
        self.s3po.upload('s3po', 'foo', 'hello')
        obj = StringIO()
        self.s3po.download('s3po', 'foo', obj)
        self.assertEqual(obj.getvalue(), 'hello')

    def test_download_filename(self):
        '''We should be able to download to a file'''
        path = os.path.join(self.tmpdir, 'foo')
        self.s3po.upload('s3po', 'foo', 'hello')
        self.s3po.download_file('s3po', 'foo', path)
        with open(path) as fin:
            self.assertEqual(fin.read(), 'hello')

    def test_upload_filename(self):
        '''We should be able to upload a file'''
        path = os.path.join(self.tmpdir, 'foo')
        with open(path, 'w+') as fout:
            fout.write('hello')
        self.s3po.upload_file('s3po', 'foo', path)
        self.assertEqual(self.s3po.download('s3po', 'foo'), 'hello')



'''Make sure our connection class works as expected'''

import unittest
from cStringIO import StringIO

from s3po.mock import Mock
from s3po.mock.key import Key, Prefix
from s3po.mock.connection import S3Connection
from s3po.mock.exceptions import S3ResponseError


class ConnectionTest(unittest.TestCase):
    '''Test our connection'''
    def setUp(self):
        self.connection = S3Connection()

    def test_lookup(self):
        '''Lookup should give us no bucket for an empty account'''
        self.assertEqual(self.connection.lookup('foo'), None)

    def test_get_bucket(self):
        '''Getting a non-existent bucket raises an error'''
        self.assertRaises(S3ResponseError, self.connection.get_bucket, 'foo')
        self.connection.create_bucket('foo')
        self.assertEqual(self.connection.get_bucket('foo').name, 'foo')

    def test_create_bucket(self):
        '''We should be able to create a bucket'''
        self.assertEqual(self.connection.lookup('foo'), None)
        self.connection.create_bucket('foo')
        self.assertEqual(self.connection.lookup('foo').name, 'foo')
        self.connection.create_bucket('foo')

    def test_get_all_buckets(self):
        '''We should be able to find all buckets'''
        self.assertEqual(self.connection.get_all_buckets(), [])
        names = ['foo', 'bar', 'whiz']
        for name in names:
            self.connection.create_bucket(name)
        self.assertEqual(set(names),
            set(b.name for b in self.connection.get_all_buckets()))

    def test_delete_bucket(self):
        '''We should be able to delete an empty bucket'''
        self.connection.create_bucket('foo')
        self.assertEqual(self.connection.get_bucket('foo').name, 'foo')
        self.connection.delete_bucket('foo')
        self.assertEqual(self.connection.lookup('foo'), None)

    def test_delete_nonempty_bucket(self):
        '''We can't delete a bucket that has keys'''
        self.connection.create_bucket('foo')
        self.connection.get_bucket('foo').new_key('bar')
        self.assertRaises(S3ResponseError, self.connection.delete_bucket, 'foo')


class BucketTest(unittest.TestCase):
    '''Test our Buckets'''
    def setUp(self):
        self.connection = S3Connection()
        self.connection.create_bucket('foo')
        self.bucket = self.connection.get_bucket('foo')

    def test_create_key(self):
        '''We should be able to make a key'''
        self.assertEqual(self.bucket.get_key('foo'), None)
        self.bucket.new_key('foo')
        self.assertEqual(self.bucket.get_key('foo').name, 'foo')

    def test_get_all_keys(self):
        '''We should be able to see all the keys'''
        self.assertEqual(self.bucket.get_all_keys(), [])
        names = ['foo', 'bar', 'whiz']
        for name in names:
            self.bucket.new_key(name)
        self.assertEqual(set(names),
            set(k.name for k in self.bucket.get_all_keys()))

    def test_delete_key(self):
        '''We should be able to delete a key'''
        self.bucket.new_key('foo')
        self.assertEqual(self.bucket.get_key('foo').name, 'foo')
        self.bucket.delete_key('foo')
        self.assertEqual(self.bucket.get_key('foo'), None)

    def test_list(self):
        '''We should be able to list all keys of a given prefix'''
        names = ['hello/foo', 'hello/bar', 'howdy/foo']
        for name in names:
            self.bucket.new_key(name)
        # All the keys with the 'hello' prefix
        self.assertEqual(set(n for n in names if n.startswith('hello')),
            set(k.name for k in self.bucket.list('hello')))
        # And now for no prefix
        self.assertEqual(set(names),
            set(k.name for k in self.bucket.list()))

    def test_delimiter_list(self):
        '''Make sure we can get keys up to a delimiter'''
        # In particular, we want to exercise that:
        #   - the delimiter appears in the prefix
        #   - a key exists before the delimiter
        #   - a key exists after the delimiter
        names = [
            'hello/howdy',
            'hello/how/are/you/today',
            'hello/how/are/you/tomorrow']
        for name in names:
            self.bucket.new_key(name)
        found = dict((k.name, k) for k in self.bucket.list('hello/', '/'))
        self.assertIsInstance(found['hello/howdy'], Key)
        self.assertIsInstance(found['hello/how'], Prefix)

    def test_missing_parts(self):
        '''Missing parts of a multi-part upload is a bad thing'''
        multi = self.bucket.initiate_multipart_upload('foo')
        multi.upload_part_from_file(StringIO('hello'), 1)
        multi.upload_part_from_file(StringIO('hello'), 2)
        multi.upload_part_from_file(StringIO('hello'), 10)
        self.assertRaises(S3ResponseError, multi.complete_upload)


class KeyTest(unittest.TestCase):
    '''Test our Key'''
    def setUp(self):
        self.connection = S3Connection()
        self.connection.create_bucket('foo')
        self.bucket = self.connection.get_bucket('foo')
        self.key = self.bucket.new_key('bar')

    def test_exists_delete(self):
        '''We should be able to verify that a key exists and delete it'''
        self.assertTrue(self.key.exists())
        self.key.delete()
        self.assertFalse(self.key.exists())

    def test_set_get_contents_as_string(self):
        '''We can set it as a string'''
        self.assertEqual(self.key.get_contents_as_string(), None)
        self.key.set_contents_from_string('hello')
        self.assertEqual(self.key.get_contents_as_string(), 'hello')

    def test_set_get_contents_as_file(self):
        '''We can set it and get it as a file'''
        self.key.set_contents_from_file(StringIO('hello'))
        obj = StringIO()
        self.key.get_contents_to_file(obj)
        self.assertEqual(obj.getvalue(), 'hello')


class FakeConn(object):
    '''An object with attributes that will be re-set when done'''
    def __init__(self):
        self.bar = 5

    def foo(self):
        '''Dummy method'''
        return 5

    def create_bucket(self, name):
        '''Create a bucket'''
        pass

    def get_bucket(self, name):
        '''Get a bucket, perhaps'''
        pass


class PatchTest(unittest.TestCase):
    '''Make sure our patching works as expected'''
    def setUp(self):
        self.conn = FakeConn()

    def test_basic(self):
        '''Make sure that it switches around all the attributes we expect'''
        self.assertEqual(self.conn.foo(), 5)
        with Mock(self.conn):
            self.assertRaises(NotImplementedError, self.conn.foo)
            self.conn.create_bucket('foo')
            self.assertEqual(self.conn.get_bucket('foo').name, 'foo')
        self.assertEqual(self.conn.foo(), 5)

    def test_exception(self):
        '''Make sure it raises an exception when appropriate'''
        self.assertEqual(self.conn.foo(), 5)

        def func():
            '''Raise an exception in the block'''
            with Mock(self.conn):
                self.assertRaises(NotImplementedError, self.conn.foo)
                self.conn.create_bucket('foo')
                self.assertEqual(self.conn.get_bucket('foo').name, 'foo')
                raise ValueError('Something went wrong')

        self.assertRaises(ValueError, func)
        self.assertEqual(self.conn.foo(), 5)



'''A Mock Key'''

from .exceptions import S3ResponseError


class Key(object):
    '''A Key in a Bucket'''
    def __init__(self, bucket, name, headers=None):
        self.bucket = bucket
        self.name = name
        self.content = None
        self.content_encoding = None
        self.headers = headers or {}
        self._size = None

    @property
    def size(self):
        '''Get the size of the content'''
        if self._size:
            return self._size
        if self.content:
            return len(self.content)
        return 0  # pragma: no cover

    def delete(self):
        '''Delete this key from S3'''
        self.bucket.delete_key(self.name)

    def exists(self):
        '''Check if this key exists'''
        return self.bucket.get_key(self.name) != None

    def get_contents_as_string(self):
        '''Get the key's contents'''
        return self.content

    def get_contents_to_file(self, fobj):
        '''Write the content out to the provided file'''
        fobj.write(self.content)

    def set_contents_from_file(self, fobj, *args, **kwargs):
        '''Set the contents based on the the file'''
        self.content = fobj.read()

    def set_contents_from_string(self, string, *args, **kwargs):
        '''Set the contents from the string'''
        self.content = string


class Prefix(object):
    '''A Prefix in a Bucket'''
    def __init__(self, bucket, name):
        self.bucket = bucket
        self.name = name


class MultiPart(object):
    '''A multipart key'''
    def __init__(self, bucket, name, headers):
        self.bucket = bucket
        self.name = name
        self.headers = headers
        self.parts = {}

    def upload_part_from_file(self, fobj, part_num):
        '''Upload another part to this file'''
        self.parts[part_num] = fobj.read()

    def complete_upload(self):
        '''Complete this as a multipart upload'''
        try:
            limits = xrange(1, len(self.parts) + 1)
            value = ''.join(self.parts[num] for num in limits)
            self.bucket.new_key(self.name).set_contents_from_string(value)
        except KeyError:
            raise S3ResponseError('Missing part from multipart upload')



'''A Mock Bucket'''

from .key import Key, Prefix, MultiPart


class Bucket(object):
    '''A Bucket in an account'''
    def __init__(self, name):
        self.name = name
        self._keys = {}

    def empty(self):
        '''Return whether or not the bucket is empty'''
        return len(self._keys) == 0

    def delete_key(self, key):
        '''Delete the provided key'''
        self._keys.pop(key, None)

    def get_all_keys(self):
        '''Get all the keys!'''
        return self._keys.values()

    def get_key(self, key):
        '''Get the provided key'''
        return self._keys.get(key.lstrip('/'), None)

    def list(self, prefix='', delimiter=''):
        '''Iterate over all the keys'''
        found = set()
        for key, value in self._keys.items():
            if key.startswith(prefix):
                # If a delimiter was provided, we should find the first
                # occurrence of it /after/ the prefix
                if delimiter:
                    location = key.find(delimiter, len(prefix) + 1)
                    if location == -1:
                        # It's a full key
                        yield value
                    else:
                        name = key[:location]
                        if name not in found:
                            found.add(name)
                            yield Prefix(self, name)
                else:
                    yield value

    def new_key(self, name):
        '''Create a new key'''
        self._keys[name] = self.get_key(name) or Key(self, name)
        return self._keys[name]

    def initiate_multipart_upload(self, name, headers=None):
        '''Start a multipart upload'''
        return MultiPart(self, name, headers)



'''Patching in our mocks'''

import inspect
from .connection import S3Connection


class Mock(object):
    '''Mock a connection object'''
    def __init__(self, conn):
        self._conn = conn
        self._mocked = S3Connection()
        self._original = {}

    def start(self):
        '''Mock all the elements of the connection'''
        callables = [(k, v) for k, v in inspect.getmembers(self._conn) if (
            not k.startswith('__') and callable(v)
        )]
        for key, value in callables:
            self._original[key] = value
            if hasattr(self._mocked, key):
                setattr(self._conn, key, getattr(self._mocked, key))
            else:
                # We need to mask that function
                def error(*args, **kwargs):
                    '''Raise an error about an attribute'''
                    raise NotImplementedError(
                        'mocks3 does not implement %s' % key)
                setattr(self._conn, key, error)

    def stop(self):
        '''Replace all the functions with their originals'''
        for key in self._original.keys():
            setattr(self._conn, key, self._original.pop(key))

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, typ, val, trace):
        self.stop()
        if typ:
            raise typ, val, trace



'''Test all of our batching utilities'''

import unittest

from s3po.connection import Connection
from s3po.exceptions import DownloadException


class BatchTest(unittest.TestCase):
    '''We should be able to batch some requests out'''
    def setUp(self):
        self.s3po = Connection()
        self.mock = self.s3po.mock()
        self.mock.start()
        self.s3po.conn.create_bucket('s3po')

    def tearDown(self):
        self.mock.stop()

    def test_upload(self):
        '''We should be able to upload'''
        names = ['foo', 'bar', 'baz', 'whiz']
        with self.s3po.batch() as batch:
            for name in names:
                batch.upload('s3po', name, 'foo')
                self.assertRaises(
                    DownloadException, self.s3po.download, 's3po', name)
        for name in names:
            self.assertEqual(self.s3po.download('s3po', name), 'foo')

    def test_callback(self):
        '''We can optionally provide a callback'''
        names = ['foo', 'bar', 'baz', 'whiz']
        with self.s3po.batch() as batch:
            for name in names:
                batch.upload('s3po', name, name)

        results = []
        with self.s3po.batch() as batch:
            for name in names:
                batch.download(
                    's3po', name, callback=lambda r: results.append(r))
        self.assertEqual(set(results), set(names))

    def test_results(self):
        '''We can get the results of our batch operations'''
        names = ['foo', 'bar', 'baz', 'whiz']
        with self.s3po.batch() as batch:
            for name in names:
                batch.upload('s3po', name, name)

        with self.s3po.batch() as batch:
            for name in names:
                batch.download('s3po', name)
        self.assertEqual(set(batch.results()), set(names))

    def test_success(self):
        '''We can get whether or not all operations returned successfully'''
        names = ['foo', 'bar', 'baz', 'whiz']
        with self.s3po.batch() as batch:
            for name in names:
                batch.upload('s3po', name, name)

        with self.s3po.batch() as batch:
            for name in names:
                batch.download('s3po', name)
        self.assertTrue(batch.success())

        # We'll change onw of the key's size so that the download fails
        self.s3po.conn.get_bucket('s3po').get_key('whiz')._size = 10
        with self.s3po.batch() as batch:
            for name in names:
                batch.download('s3po', name, retries=0)
        self.assertFalse(batch.success())



'''Connecting to and talking to S3'''

import os
import warnings
from cStringIO import StringIO
from boto.s3.connection import S3Connection

# Internal imports
from .mock import Mock
from .util import retry
from .util import CountFile, logger
from .exceptions import UploadException, DownloadException


class Connection(object):
    '''Our connection to S3'''
    # How big must a file get before it's multiparted. Also how big the chunks
    # are that we'll read
    multipart_chunk = 50 * 1024 * 1204

    def __init__(self, *args, **kwargs):
        self.conn = S3Connection(*args, **kwargs)

    def batch(self, poolsize=20):
        from .batch import Batch
        return Batch(self, poolsize)

    def mock(self):
        '''Return a context-manager for managing our mocking'''
        return Mock(self.conn)

    def _download(self, bucket, key, fobj, retries):
        '''Download the contents of bucket/key to fobj'''
        bucket = self.conn.get_bucket(bucket)
        # Make a file that we'll write into
        fobj = CountFile(fobj)
        obj = bucket.get_key(key)
        if not obj:
            raise DownloadException('Key %s does not exist in %s' % (
                key, bucket.name))

        # Get its original location so we can go back to it if need be
        offset = fobj.tell()

        @retry(retries)
        def func():
            '''The bit that we want to retry'''
            fobj.seek(offset)
            obj.get_contents_to_file(fobj)
            # Ensure it was downloaded completely
            logger.info(
                'Downloaded %s bytes out of %s' % (fobj.count, obj.size))
            if obj.size != fobj.count:
                raise DownloadException('Downloaded only %i of %i bytes' % (
                    fobj.count, obj.size or 0))
        # With our wrapped function defined, we'll go ahead an invoke it.
        func()

    def _upload(self, bucket, key, fobj, retries, headers=None):
        '''Upload the contents of fobj to bucket/key with headers'''
        # Make our headers object
        headers = headers or {}
        bucket = self.conn.get_bucket(bucket)
        # We'll read in some data, and if the file appears small enough, we'll
        # upload it in a single go. In order for it to be a valid multipart
        # upload, it needs at least two parts, so we will make sure there are
        # at least enough for two parts before we commit to multipart
        data = fobj.read(2 * self.multipart_chunk)
        if len(data) < (2 * self.multipart_chunk):
            key = bucket.new_key(key)

            @retry(retries)
            def func():
                '''The bit that we want to retry'''
                key.set_contents_from_string(data, headers=headers)
                if key.size != len(data):
                    raise UploadException('Uploaded only %i for %i bytes' % (
                        key.size, len(data)))
                return True
            return func()
        else:
            logger.info('Multipart')
            # Otherwise, it's a large-enough file that we should multipart
            # upload it. There's a restriction that all parts of a multipart
            # upload must be at least 5MB. Therefore, we should keep uploading
            # chunks as long as the remaining data is 5MB greater than our chunk
            # size. That way we avoid the case where we have a remainder less
            # than this limit
            multi = bucket.initiate_multipart_upload(key, headers=headers)
            count = 1
            while len(data) >= (self.multipart_chunk + (5 * 1024 * 1024)):
                part = data[0:self.multipart_chunk]
                retry(
                    retries)(multi.upload_part_from_file)(StringIO(part), count)
                data = (
                    data[self.multipart_chunk:] +
                    fobj.read(self.multipart_chunk))
                count += 1
            # And finally, the last part
            multi.upload_part_from_file(StringIO(data), count)
            multi.complete_upload()
            return True

    def uploadFile(self, bucket, key, path, headers=None, compress=None,
        retries=3, async=None, delete=None, silent=False):  # pragma: no cover
        '''Upload the file at path to bucket/key'''
        warnings.warn('Deprecation: use `upload` instead of `uploadFile`',
            DeprecationWarning)
        with open(os.path.abspath(path)) as fobj:
            return self.upload(bucket, key, fobj, headers, retries)

    def uploadString(self, bucket, key, string, headers=None, compress=None,
        retries=3, async=None, silent=False):  # pragma: no cover
        '''Upload the string to bucket/key'''
        warnings.warn('Deprecation: use `upload` instead of `uploadString`',
            DeprecationWarning)
        return self.upload(bucket, key, string, headers, retries)

    def downloadFile(self, bucket, key, filename, retries=3): # pragma: no cover
        '''Download bucket/key to the provided filename'''
        warnings.warn('Deprecation: use `download` instead of `downloadFile`',
            DeprecationWarning)
        with open(filename, 'w') as fout:
            return self.download(bucket, key, fout, retries)

    def downloadString(self, bucket, key, retries=3): # pragma: no cover
        '''Download bucket/key and return the string'''
        warnings.warn('Deprecation: use `download` instead of `downloadString`',
            DeprecationWarning)
        return self.download(bucket, key, retries=retries)

    def upload(self, bucket, key, obj_or_data, headers=None, retries=3):
        '''Upload the provided string or file object to bucket/key'''
        logger.info('Uploading to %s / %s' % (bucket, key))
        if isinstance(obj_or_data, basestring):
            return self._upload(
                bucket, key, StringIO(obj_or_data), retries, headers)
        else:
            return self._upload(bucket, key, obj_or_data, retries, headers)

    def upload_file(self, bucket, key, path, headers=None, retries=3):
        '''Upload the file at path to bucket/key. This method is important for
        use in batch mode, so that the file object can be used with the right
        context management'''
        with open(os.path.abspath(path)) as fobj:
            return self.upload(bucket, key, fobj, headers, retries)

    def download(self, bucket, key, obj=None, retries=3):
        '''Download to either the object or return a string'''
        if obj:
            return self._download(bucket, key, obj, retries)
        obj = StringIO()
        self._download(bucket, key, obj, retries)
        return obj.getvalue()

    def download_file(self, bucket, key, path, retries=3, mode='w'):
        '''Download the item at bucket/key to a file at path. This method is
        important for us in batch mode so that the file object can be used with
        the right context management'''
        with open(os.path.abspath(path), mode) as fout:
            return self.download(bucket, key, fout, retries)
